
#include "FFmpegPlayer.h"


FFmpegPlayer::FFmpegPlayer(std::string filename, int openCVType)
    : m_filename( filename )
    , m_openCVType( openCVType )
    , m_formatCtx( nullptr )
{
    m_bpp = cv::Mat(1,1,openCVType).elemSize() * 8;
}

FFmpegPlayer::~FFmpegPlayer()
{
    Close();
}

int FFmpegPlayer::decode_packet( cv::Mat& mat)
{
    int ret = 0;
    int decoded = m_packet.size;

    /* decode video frame */
    ret = avcodec_send_packet(m_codecCtx, &m_packet);
    if (ret < 0) {
        return ret;
    }

    ret = avcodec_receive_frame(m_codecCtx, m_frame);
    if (ret == 0) {

        if (m_frame->width != m_width || m_frame->height != m_height ||
            m_frame->format != m_pixelFormat) {
            /* To handle this change, one could call av_image_alloc again and
             * decode the following frames into another rawvideo file. */
            fprintf(stderr, "Error: Width, height and pixel format have to be "
                    "constant in a rawvideo file, but the width, height or "
                    "pixel format of the input video changed:\n"
                    "old: width = %d, height = %d, format = %s\n"
                    "new: width = %d, height = %d, format = %s\n",
                    m_width, m_height, av_get_pix_fmt_name(m_pixelFormat),
                    m_frame->width, m_frame->height,
                    av_get_pix_fmt_name((AVPixelFormat)m_frame->format));
            return -1;
        }

        /* copy decoded frame to destination buffer:
         * this is required since rawvideo expects non aligned data */
        av_image_copy(m_videoData, m_videoLinesize,
                      (const uint8_t **)(m_frame->data), m_frame->linesize,
                      m_pixelFormat, m_width, m_height);

        /* write to mat */
        mat = cv::Mat( m_height, m_width / ( m_bpp / 16 ), m_openCVType, m_videoData[0] );
    }

    return decoded;
}

int FFmpegPlayer::open_codec_context(int *stream_idx)
{
    int ret, stream_index;
    AVStream *st;
    AVCodec *dec = NULL;
    AVDictionary *opts = NULL;

    ret = av_find_best_stream(m_formatCtx, AVMEDIA_TYPE_VIDEO, -1, -1, NULL, 0);
    if (ret < 0) {
        fprintf(stderr, "Could not find %s stream in input file '%s'\n",
                av_get_media_type_string(AVMEDIA_TYPE_VIDEO), m_filename.c_str());
        return ret;
    } else {
        stream_index = ret;
        st = m_formatCtx->streams[stream_index];

        /* find decoder for the stream */
        dec = avcodec_find_decoder(st->codecpar->codec_id);
        if (!dec) {
            fprintf(stderr, "Failed to find %s codec\n",
                    av_get_media_type_string(AVMEDIA_TYPE_VIDEO));
            return AVERROR(EINVAL);
        }

        /* Allocate a codec context for the decoder */
        m_codecCtx = avcodec_alloc_context3(dec);
        if (!m_codecCtx) {
            fprintf(stderr, "Failed to allocate the %s codec context\n",
                    av_get_media_type_string(AVMEDIA_TYPE_VIDEO));
            return AVERROR(ENOMEM);
        }

        /* Copy codec parameters from input stream to output codec context */
        if ((ret = avcodec_parameters_to_context(m_codecCtx, st->codecpar)) < 0) {
            fprintf(stderr, "Failed to copy %s codec parameters to decoder context\n",
                    av_get_media_type_string(AVMEDIA_TYPE_VIDEO));
            return ret;
        }

        /* Init the decoders, with or without reference counting */
        av_dict_set(&opts, "refcounted_frames", "0", 0);
        if ((ret = avcodec_open2(m_codecCtx, dec, &opts)) < 0) {
            fprintf(stderr, "Failed to open %s codec\n",
                    av_get_media_type_string(AVMEDIA_TYPE_VIDEO));
            return ret;
        }
        *stream_idx = stream_index;
    }

    return 0;
}

bool FFmpegPlayer::Open()
{
#if LIBAVCODEC_VERSION_INT < AV_VERSION_INT(58, 9, 100)
    avcodec_register_all();
#endif
    av_init_packet( &m_packet );

    int ret = 0;

    /* open input file, and allocate format context */
    if (avformat_open_input(&m_formatCtx, m_filename.c_str(), NULL, NULL) < 0) {
        fprintf(stderr, "Could not open source file %s\n", m_filename.c_str());
        return false;
    }

    /* retrieve stream information */
    if (avformat_find_stream_info(m_formatCtx, NULL) < 0) {
        fprintf(stderr, "Could not find stream information\n");
        return false;
    }

    int streamIdx;
    if (open_codec_context(&streamIdx) >= 0) {
        m_stream = m_formatCtx->streams[streamIdx];

        /* allocate image where the decoded image will be put */
        m_width = m_codecCtx->width;
        m_height = m_codecCtx->height;
        m_pixelFormat = m_codecCtx->pix_fmt;
        ret = av_image_alloc(m_videoData, m_videoLinesize, m_width, m_height, m_pixelFormat, 1);
        if (ret < 0) {
            fprintf(stderr, "Could not allocate raw video buffer\n");
            return false;
        }
        m_videoBufferSize = ret;
    }

    /* dump input information to stderr */
    av_dump_format(m_formatCtx, 0, m_filename.c_str(), 0);

    m_frame = av_frame_alloc();
    if (!m_frame) {
        fprintf(stderr, "Could not allocate frame\n");
        ret = AVERROR(ENOMEM);
        return false;
    }

    /* initialize packet, set data to NULL, let the demuxer fill it */
    av_init_packet(&m_packet);
    m_packet.data = NULL;
    m_packet.size = 0;

    return true;
}

double FFmpegPlayer::processFrame( cv::Mat& out )
{
    int ret = 0;

    if( m_packet.size <= 0 )
        ret = av_read_frame(m_formatCtx, &m_packet);

    if (ret >= 0) {
        AVPacket orig_pkt = m_packet;
        ret = decode_packet( out );
        if (ret < 0)
            return -1;

        m_packet.data += ret;
        m_packet.size -= ret;

        if( m_packet.size <= 0 )
            av_packet_unref(&orig_pkt);

        return m_codecCtx->ticks_per_frame * ( (double)m_codecCtx->time_base.num / (double)m_codecCtx->time_base.den );
    }
    else
        return -1;
}

void FFmpegPlayer::Close()
{
    avcodec_close( m_codecCtx );
}

FFmpegCamPlayer::FFmpegCamPlayer(std::string filePrefix, bool playInRealtime)
    : m_playInRealtime( playInRealtime )
{
    /*std::ifstream metaDataFile( filePrefix + "_metadata.txt" );

    while( metaDataFile.is_open() && !metaDataFile.eof() )
    {
        StreamTypes stream;
        cv::Size resolution;
        int cvPixelType = -1;
        CamIntrinsics intrinsics;

        metaDataFile >> (unsigned int&)stream >> resolution.width >> resolution.height >> cvPixelType;
        metaDataFile >> cx >> cy >> fx >> fy
                     >> k1 >> k2 >> k3 >> p1 >> p2;

        if( cvPixelType >= 0 )
        {
            m_streams.insert( stream );
            m_streamPlayers.emplace( std::piecewise_construct, std::make_tuple(stream),
                                     std::make_tuple(filePrefix + "_" + std::to_string((unsigned int)stream) + ".h264", cvPixelType) );
            ImageStreams.emplace( std::piecewise_construct, std::make_tuple(stream),
                                  std::make_tuple(resolution, cvPixelType, MakeVar<D>(intrinsics)) );
            m_eventSources.emplace( stream, MakeEventSource<D, cv::Mat>());
            ImageStreams[stream].events = m_eventSources[stream];
        }
    }*/
	std::ifstream metaDataFile(filePrefix + "_metadata.txt");
	std::stringstream os;
	if (metaDataFile.is_open()) {
		os << metaDataFile.rdbuf();
		metaDataFile.close();
		cereal::JSONInputArchive archive_in(os);
		archive_in(Streams);
		std::vector<std::shared_ptr<Stream>> allStreams = Streams.GetAllStreams();
		for (auto stream : allStreams)
		{
			auto imgStream = std::dynamic_pointer_cast<ImageStream>(stream);
			if (imgStream && imgStream->m_cvPixelType.Value() >= 0)
			{
				m_streams.insert(imgStream->getType());
				m_streamPlayers.emplace(std::piecewise_construct, std::make_tuple(imgStream->getType()),
					std::make_tuple(filePrefix + "_" + std::to_string((unsigned int)imgStream->getType()) + ".h264", imgStream->m_cvPixelType.Value()));
				m_eventSources.emplace(imgStream->getType(), MakeEventSource<D, cv::Mat>());
			}
		}
	}
}

FFmpegCamPlayer::~FFmpegCamPlayer()
{
    Close();
}

bool FFmpegCamPlayer::Open()
{
    for( auto& player : m_streamPlayers )
        if( !player.second.Open() )
            return false;


    doProcessing.store(true);
    processingThread = std::thread( [this] ()
    {
        std::vector<cv::Mat> mats( m_streamPlayers.size() );
        double pts = 0;

        auto start = std::chrono::high_resolution_clock::now();

        while( doProcessing )
        {
            unsigned int i=0;
            for( auto& player : m_streamPlayers )
                pts = player.second.processFrame( mats[i++] );

            if( pts > 0 && m_playInRealtime )
            {
                auto now = std::chrono::high_resolution_clock::now();
                auto secsFromLast = std::chrono::duration<double>( now - start );
                start = now;
                if( secsFromLast.count() < pts )
                    std::this_thread::sleep_for( std::chrono::duration<double>( pts - secsFromLast.count() ) );
            }

            i=0;
            for( auto& player : m_streamPlayers )
                m_eventSources[player.first] << mats[i++];
        }

        doProcessing.store(false);
    });

    return true;
}

void FFmpegCamPlayer::Close()
{
    doProcessing.store(false);
    if( processingThread.joinable() )
        processingThread.join();

    for( auto& player : m_streamPlayers )
        player.second.Close();
}

std::set<StreamTypes> FFmpegCamPlayer::getStreamCapabilities()
{
    return m_streams;
}
