
#include "PatternTracker.h"

#include <memory>
#include <src/Core/ImageStream.h>
#include <src/Core/IntrinsicsStream.h>

PatternTracker::PatternTracker(StreamCollection streams, cv::Size patternSize, double patternSpacing)
    : m_patternSize( patternSize )
    , m_patternSpacing( patternSpacing )
    , streams_(streams)
    , m_lastPatternFound( std::chrono::high_resolution_clock::now() )
{
    // calculate the corner positions on the pattern
    for( int i = 0; i < m_patternSize.height; i++ )
        for( int j = 0; j < m_patternSize.width; j++ )
            m_corners.emplace_back((2*j + i % 2)*m_patternSpacing, i*m_patternSpacing, 0);

    intrFileName_ = "camIntrinsics_" + streams_.GetSerial() + ".dat";
}

bool PatternTracker::start()
{
    auto irStream = streams_.GetStreamFirst<ImageStream>(StreamTypes::IR);
    assert(irStream);

    // use intrinsics from collection or add them
    auto irIntrinsics = streams_.GetStreamFirst<IntrinsicsStream>(StreamTypes::INTRINSICS, irStream->GetId());

    // try to read from file if no intrinsic were found
    if (!irIntrinsics) {
        std::ifstream file(intrFileName_.c_str());
        if (file.good()) {
            auto intr = CamIntrinsics(intrFileName_);
            auto irIntrStream = std::make_shared<IntrinsicsStream>(irStream->GetId());
            irIntrinsics->Intrinsics.Set(intr);
            streams_.AddStream(irIntrStream);
        }
    }

    // do we have intrinsics now?
    if (irIntrinsics) {
        m_patternCollectionDone.Set( true );
        m_intrinsicCalibrationDone.Set( true );
    }

    // collect patterns from the IR image and undistort it when the intrinsic calibration was done
    auto irUndistorted =
    Transform(irStream->events, [&](cv::Mat irImage) {
        double min;
        double max;
        cv::minMaxIdx(irImage, &min, &max);

        cv::Mat pat;
        irImage.convertTo(pat, CV_8UC1, 255.0 / (max - min), -min);

        auto timeSinceLastPattern = std::chrono::duration<double>(std::chrono::high_resolution_clock::now() - m_lastPatternFound);
        if( !m_patternCollectionDone.Value() && timeSinceLastPattern.count() > 0.5 )
        {
            std::vector<cv::Point2f> centers;
            bool patternfound = cv::findCirclesGrid( pat, m_patternSize, centers, cv::CALIB_CB_ASYMMETRIC_GRID );
            cv::drawChessboardCorners(pat, m_patternSize, cv::Mat(centers), patternfound);

            if( patternfound )
            {
                m_lastPatternFound = std::chrono::high_resolution_clock::now();
                m_allCenters.push_back( centers );
                if( m_allCenters.size() >= 20 ) {
                    AsyncTransaction<D>([this] {
                        m_patternCollectionDone.Set(true);
                    });
                }

                std::cout << "Found pattern #" << m_allCenters.size() << std::endl;
            }
        }

        cv::Mat ret;
        if( m_intrinsicCalibrationDone.Value() )
            cv::undistort( pat, ret, m_camMat, m_distCoeffs, m_newCamMat ); 
        else
            ret = pat;

        return ret;
    });

    // Once we have enough patterns do the callibration
    m_calibrator =
    Observe( m_patternCollectionDone, [&] (bool done) {
        if( done )
        {
            // * calibrate the camera
            // create the object points vector
            std::vector<cv::Point3f> corners;
            for( int i = 0; i < m_patternSize.height; i++ )
                for( int j = 0; j < m_patternSize.width; j++ )
                    corners.push_back( cv::Point3f((2*j + i % 2)*20, i*20, 0 ));

            std::vector<std::vector<cv::Point3f>> objPoints( m_allCenters.size(), corners );

            // calibrate
            cv::Size camResolution = irStream->m_resolution.Value();
            std::vector<cv::Mat> rvecs, tvecs;
            calibrateCamera( objPoints, m_allCenters, camResolution, m_camMat, m_distCoeffs, rvecs, tvecs );

            // compute the reprojection error
            std::vector<float> reprojErrs;
            double totalAvgErr = computeReprojectionErrors(objPoints, m_allCenters, rvecs, tvecs,
                                                           m_camMat, m_distCoeffs, reprojErrs);

            std::cout << "Total average reprojection error: " << totalAvgErr << std::endl;

            // compute the optimal, zoomed and cropped camera mat
            m_newCamMat = cv::getOptimalNewCameraMatrix( m_camMat, m_distCoeffs, camResolution, 0 );

            // add intrinsics to stream
            CamIntrinsics intr(m_camMat, m_distCoeffs, m_newCamMat);
            auto irIntrStream = std::make_shared<IntrinsicsStream>(irStream->GetId());
            irIntrinsics->Intrinsics.Set(intr);
            streams_.AddStream(irIntrStream);
            // TODO: replace with cereal serialization
            intr.WriteToFile(intrFileName_);

            AsyncTransaction<D>([this] {
                m_intrinsicCalibrationDone.Set( true );
            });
        }
    });

    // retrieve the camera extrinsic transformation matrix whenever a pattern was found
    cv::namedWindow(streams_.GetSerial() + " tracked pattern", cv::WINDOW_NORMAL);
    m_coordSystemStream =
    Transform( irUndistorted, [this](cv::Mat undistortedIR) {

        cv::Matx44d transMat = cv::Matx44d::eye();

        if( m_intrinsicCalibrationDone.Value() )
        {
            cv::Mat viewColor;
            cvtColor(undistortedIR, viewColor, cv::COLOR_GRAY2BGR);

            std::vector<cv::Point2f> centers;
            bool patternfound = cv::findCirclesGrid( undistortedIR, m_patternSize, centers, cv::CALIB_CB_ASYMMETRIC_GRID );

            if( patternfound )
            {
                if( !cv::solvePnPRansac( m_corners, centers, m_camMat, m_distCoeffs, m_rvec, m_tvec, true ) )
                    return transMat;

                // calculate the rotation matrix and put it in the 4x4 transformation matrix
                cv::Matx33d rotMat;
                cv::Matx44d rotMatHom = cv::Matx44d::eye();

                cv::Rodrigues( m_rvec, rotMat );
                for( int i = 0; i < 3; ++i )
                    for( int j = 0; j < 3; ++j )
                        rotMatHom(i,j) = rotMat(i,j);

                cv::Matx44d transMat2 = cv::Matx44d::eye();
                // put the translation vector into the transformation matrix
                for( int i = 0; i < 3; ++i )
                    transMat2(i,3) = m_tvec.at<double>(i);

                transMat =  transMat2 * rotMatHom;

                // DEBUG: project axes on the IR image plane and show them in the IR image
                std::vector<cv::Point2f> projectedPoints;
                std::vector<cv::Point3f> axis;
                const float l = m_patternSpacing * 3.f;
                axis.push_back(cv::Point3f(l,0,0));
                axis.push_back(cv::Point3f(0,l,0));
                axis.push_back(cv::Point3f(0,0,l));

                cv::projectPoints( axis, m_rvec, m_tvec, m_camMat, m_distCoeffs, projectedPoints );

                cv::line(viewColor, centers.at(0), projectedPoints.at(0), cv::Scalar(0,0,255), 2);
                cv::line(viewColor, centers.at(0), projectedPoints.at(1), cv::Scalar(0,255,0), 2);
                cv::line(viewColor, centers.at(0), projectedPoints.at(2), cv::Scalar(255,0,0), 2);
            }

            cv::imshow(streams_.GetSerial() + " tracked pattern", viewColor);
        }
        else
            cv::imshow(streams_.GetSerial() + " tracked pattern", undistortedIR);

        return transMat;
    });

    return true;
}

double PatternTracker::computeReprojectionErrors(const std::vector<std::vector<cv::Point3f> >& objectPoints,
                                                 const std::vector<std::vector<cv::Point2f> >& imagePoints,
                                                 const std::vector<cv::Mat>& rvecs, const std::vector<cv::Mat>& tvecs,
                                                 const cv::Mat& cameraMatrix, const cv::Mat& distCoeffs,
                                                 std::vector<float>& perViewErrors )
{
    std::vector<cv::Point2f> imagePoints2;
    int i, totalPoints = 0;
    double totalErr = 0, err;
    perViewErrors.resize(objectPoints.size());

    for( i = 0; i < (int)objectPoints.size(); i++ )
    {
        projectPoints(cv::Mat(objectPoints[i]), rvecs[i], tvecs[i],
                      cameraMatrix, distCoeffs, imagePoints2);
        err = norm(cv::Mat(imagePoints[i]), cv::Mat(imagePoints2), cv::NORM_L2);
        int n = (int)objectPoints[i].size();
        perViewErrors[i] = (float)std::sqrt(err*err/n);
        totalErr += err*err;
        totalPoints += n;
    }

    return std::sqrt(totalErr/totalPoints);
}


