#include "UnrelInterface.h"

#include <memory>
#include "src/Cams/Freenect2/Freenect2Cam.h"
#include "src/Cams/NetCam/StreamClient.h"
#include "src/Core/ImageStream.h"
#include "src/Cams/Netcam/CompressedStream.h"
#include "src/Core/IntrinsicsStream.h"
#include "src/Compression/Lz4Compression.h"
#include "src/Compression/Lz4QuantCompression.h"
#include "src/Compression/QuantizedCompression.h"
#include "src/Compression/ColorCompression.h"
#include "src/Reconstruction/CloudFromDepth.h"

class UnrealInterface::impl {  
public:
    //std::shared_ptr<BaseCam> cam;
    cv::Mat lastColor = cv::Mat(424, 512, CV_8UC3);
    cv::Mat lastCloud = cv::Mat(424 * 512, 3, CV_32F);
    CloudFromDepth* m_pointsMapper;

    std::mutex lockColor;
    std::mutex lockCloud;

    // todo make atomic?
    bool dirtyColor = false;
    bool dirtyCloud = false;

	std::string ip;
	std::string port;

	std::vector<std::unique_ptr<dyncam::StreamClient>> m_netClients;
};

// we need this here for the pimpl
UnrealInterface::~UnrealInterface() = default;

UnrealInterface::UnrealInterface() : pimpl(new impl)
{

}

uint8_t* UnrealInterface::getPCColor() const
{
    std::lock_guard<std::mutex> lock(this->pimpl->lockColor);
    pimpl->dirtyColor = false;
    return reinterpret_cast<uint8_t*>(pimpl->lastColor.data);
}

float* UnrealInterface::getPCPoints() const
{
    std::lock_guard<std::mutex> lock(this->pimpl->lockCloud);
    pimpl->dirtyCloud = false;
	
    return reinterpret_cast<float*>(pimpl->lastCloud.data);
}

bool UnrealInterface::getPointsDirty() const
{
	return pimpl->dirtyCloud && pimpl->dirtyColor;
}

void UnrealInterface::setCout(std::streambuf * sb)
{
    std::cout.rdbuf(sb);
}

void UnrealInterface::initConnection(const char * IPIn, const char * portIn)
{

	//std::cout << "UEClient: STRINGS:"  << IPIn << "   " << portIn << std::endl;
	// TODO: check validity of IP and port
	std::cout << "UEClient: Start to init connetion" << std::endl;
	pimpl->ip = std::string(IPIn);
	pimpl->port = std::string(portIn);
	auto client = std::make_unique<StreamClient>();
	client->start(pimpl->ip, pimpl->port);
	pimpl->m_netClients.push_back(std::move(client));
}

void UnrealInterface::killConnections()
{
	for (int i = 0; i < pimpl->m_netClients.size(); ++i)
	{
		if (pimpl->m_netClients.at(i))
		{
			std::cout << "UEClient: Killing Connetion" << i << std::endl;
			pimpl->m_netClients.at(i).reset();
		}
	}
	pimpl->m_netClients.clear();
}

bool UnrealInterface::requestStreams()
{
	if (!pimpl->m_netClients.front())
	{
		std::cout << "UEClient: Could not find client!" << std::endl;
		return false;
	}

	auto &streams = pimpl->m_netClients.front()->m_streams;

	std::shared_ptr<CompressedStream> cs_streamRC = std::make_shared<CompressedStream>(streams->GetStreamFirst<ImageStream>(StreamTypes::REGISTERED_COLOR),
		std::make_unique<ColorCompression>());

	std::shared_ptr<CompressedStream> cs_streamC = std::make_shared<CompressedStream>(streams->GetStreamFirst<ImageStream>(StreamTypes::COLOR),
		std::make_unique<ColorCompression>());

	std::shared_ptr<CompressedStream> cs_streamUD = std::make_shared<CompressedStream>(streams->GetStreamFirst<ImageStream>(StreamTypes::UNDISTORTED_DEPTH),
		std::make_unique<Lz4QuantCompression>(cv::Size(512, 424), CV_16S));
	
	if (!cs_streamRC->m_innerStream || !cs_streamC->m_innerStream)
	{
		std::cout << "UEClient: Could not find streams to request!" << std::endl;
		return false;
	}

	std::cout << "UEClient: Requesting streams" << std::endl;

	pimpl->m_netClients.front()->requestStream(cs_streamRC);
	//pimpl->m_netClients.front()->requestStream(cs_streamC);
	pimpl->m_netClients.front()->requestStream(cs_streamUD);
	pimpl->m_netClients.front()->requestStream(streams->GetStreamFirst<ImageStream>(StreamTypes::IR));
	//pimpl->m_netClients.front()->requestStream(streams->GetStreamFirst<ImageStream>(StreamTypes::UNDISTORTED_DEPTH));
	//pimpl->m_netClients.front()->requestStream(streams->GetStreamFirst<ImageStream>(StreamTypes::REGISTERED_COLOR));
	//pimpl->m_netClients.front()->requestStream(streams->GetStreamFirst<ImageStream>(StreamTypes::DEPTH));
	pimpl->m_netClients.front()->requestStream(streams->GetStreamFirst<IntrinsicsStream>(StreamTypes::INTRINSICS));

	pimpl->m_pointsMapper = new CloudFromDepth(*pimpl->m_netClients.front()->m_streams);

	Observe(this->pimpl->m_pointsMapper->PointPosEvents, [this](const cv::Mat &points) {
		std::lock_guard<std::mutex> lock(this->pimpl->lockCloud);

		assert(this->pimpl->lastCloud.dims == points.dims);
		assert(this->pimpl->lastCloud.size == points.size);
		assert(this->pimpl->lastCloud.elemSize() == points.elemSize());

		this->pimpl->lastCloud = points.clone();
		this->pimpl->dirtyCloud = true;
	});

	Observe(this->pimpl->m_pointsMapper->PointColorsEvents, [this](const cv::Mat &colors) {
		std::lock_guard<std::mutex> lock(this->pimpl->lockColor);

		assert(this->pimpl->lastColor.dims == colors.dims);
		assert(this->pimpl->lastColor.size == colors.size);
		assert(this->pimpl->lastColor.elemSize() == colors.elemSize());

		this->pimpl->lastColor = colors.clone();
		this->pimpl->dirtyColor = true;
	});
	return true;
}
